@@ -49,7 +49,7 @@ from getmailcore.constants import *
 from getmailcore.message import *
 from getmailcore.utilities import *
 from getmailcore._pop3ssl import POP3SSL, POP3_ssl_port
-from getmailcore.baseclasses import ConfigurableBase
+from getmailcore.baseclasses import *
 
 NOT_ENVELOPE_RECIPIENT_HEADERS = (
     'to',
@@ -94,27 +94,15 @@ class Py24POP3SSLinitMixIn(object):
         if not hasattr(socket, 'ssl'):
             raise getmailConfigurationError('SSL not supported by'
                 ' this installation of Python')
-        if (self.conf['keyfile'] is not None
-                and not os.path.isfile(self.conf['keyfile'])):
-            raise getmailConfigurationError('optional keyfile must be'
-                ' path to a valid file')
-        if (self.conf['certfile'] is not None
-                and not os.path.isfile(self.conf['certfile'])):
-            raise getmailConfigurationError('optional certfile must be'
-                ' path to a valid file')
-        if not (self.conf['certfile'] == self.conf['keyfile'] == None):
-            if self.conf['certfile'] is None or self.conf['keyfile'] is None:
-                raise getmailConfigurationError('optional certfile and keyfile'
-                    ' must be supplied together')
+        (keyfile, certfile) = check_ssl_key_and_cert(self.conf)
         try:
-            if self.conf['certfile'] and self.conf['keyfile']:
+            if keyfile:
                 self.log.trace('establishing POP3 SSL connection to %s:%d'
                     ' with keyfile %s, certfile %s'
-                    % (self.conf['server'], self.conf['port'],
-                        self.conf['keyfile'], self.conf['certfile'])
+                    % (self.conf['server'], self.conf['port'], keyfile, certfile)
                     + os.linesep)
                 self.conn = poplib.POP3_SSL(self.conf['server'], self.conf['port'],
-                    self.conf['keyfile'], self.conf['certfile'])
+                    keyfile, certfile)
             else:
                 self.log.trace('establishing POP3 SSL connection to %s:%d'
                     % (self.conf['server'], self.conf['port'])
@@ -128,9 +116,9 @@ class Py24POP3SSLinitMixIn(object):
             raise getmailOperationError('error resolving name %s during'
                 ' connect (%s)' % (self.conf['server'], o))
 
-	self.conn.sock.setblocking(1)
+        self.conn.sock.setblocking(1)
         self.log.trace('POP3 connection %s established' % self.conn 
-            + os.linesep)
+                       + os.linesep)
 
 #######################################
 class Py23POP3SSLinitMixIn(object):
@@ -142,27 +130,16 @@ class Py23POP3SSLinitMixIn(object):
         if not hasattr(socket, 'ssl'):
             raise getmailConfigurationError('SSL not supported by'
                 ' this installation of Python')
-        if (self.conf['keyfile'] is not None
-                and not os.path.isfile(self.conf['keyfile'])):
-            raise getmailConfigurationError('optional keyfile must be'
-                ' path to a valid file')
-        if (self.conf['certfile'] is not None
-                and not os.path.isfile(self.conf['certfile'])):
-            raise getmailConfigurationError('optional certfile must be'
-                ' path to a valid file')
-        if not (self.conf['certfile'] == self.conf['keyfile'] == None):
-            if self.conf['certfile'] is None or self.conf['keyfile'] is None:
-                raise getmailConfigurationError('optional certfile and keyfile'
-                    ' must be supplied together')
+        (keyfile, certfile) = check_ssl_key_and_cert(self.conf)
         try:
-            if self.conf['certfile'] and self.conf['keyfile']:
+            if keyfile:
                 self.log.trace('establishing POP3 SSL connection to %s:%d'
                     ' with keyfile %s, certfile %s'
                     % (self.conf['server'], self.conf['port'],
-                        self.conf['keyfile'], self.conf['certfile'])
+                       keyfile, certfile)
                     + os.linesep)
                 self.conn = POP3SSL(self.conf['server'], self.conf['port'],
-                    self.conf['keyfile'], self.conf['certfile'])
+                    keyfile, certfile)
             else:
                 self.log.trace('establishing POP3 SSL connection to %s:%d'
                     % (self.conf['server'], self.conf['port'])
@@ -207,28 +184,16 @@ class IMAPSSLinitMixIn(object):
         if not hasattr(socket, 'ssl'):
             raise getmailConfigurationError('SSL not supported by'
                 ' this installation of Python')
-        if (self.conf['keyfile'] is not None
-                and not os.path.isfile(self.conf['keyfile'])):
-            raise getmailConfigurationError('optional keyfile must be'
-                ' path to a valid file')
-        if (self.conf['certfile'] is not None
-                and not os.path.isfile(self.conf['certfile'])):
-            raise getmailConfigurationError('optional certfile must be'
-                ' path to a valid file')
-        if not (self.conf['certfile'] == self.conf['keyfile'] == None):
-            if self.conf['certfile'] is None or self.conf['keyfile'] is None:
-                raise getmailConfigurationError('optional certfile and keyfile'
-                    ' must be supplied together')
+        (keyfile, certfile) = check_ssl_key_and_cert(self.conf)
         try:
-            if self.conf['certfile'] and self.conf['keyfile']:
+            if keyfile:
                 self.log.trace('establishing IMAP SSL connection to %s:%d'
                     ' with keyfile %s, certfile %s'
                     % (self.conf['server'], self.conf['port'],
-                        self.conf['keyfile'], self.conf['certfile'])
+                       keyfile, certfile)
                     + os.linesep)
                 self.conn = imaplib.IMAP4_SSL(self.conf['server'],
-                    self.conf['port'], self.conf['keyfile'],
-                    self.conf['certfile'])
+                    self.conf['port'], keyfile, certfile)
             else:
                 self.log.trace('establishing IMAP SSL connection to %s:%d'
                     % (self.conf['server'], self.conf['port']) + os.linesep)
@@ -268,7 +233,8 @@ class RetrieverSkeleton(ConfigurableBase):
       __str__(self) - return a simple string representing the class instance.
 
       _getmsglist(self) - retieve a list of all available messages, and store
-                          unique message identifiers in the list self.msgids.
+                          unique message identifiers in the dict 
+                          self.msgnum_by_msgid.
                           Message identifiers must be unique and persistent
                           across instantiations.  Also store message sizes (in
                           octets) in a dictionary self.msgsizes, using the
@@ -305,8 +271,8 @@ class RetrieverSkeleton(ConfigurableBase):
     '''
 
     def __init__(self, **args):
-        ConfigurableBase.__init__(self, **args)
-        self.msgids = []
+        self.msgnum_by_msgid = {}
+        self.msgid_by_msgnum = {}
         self.msgsizes = {}
         self.headercache = {}
         self.oldmail = {}
@@ -316,6 +282,7 @@ class RetrieverSkeleton(ConfigurableBase):
         self.__oldmail_written = False
         self.__initialized = False
         self.gotmsglist = False
+        ConfigurableBase.__init__(self, **args)
 
     def __del__(self):
         self.log.trace()
@@ -327,13 +294,13 @@ class RetrieverSkeleton(ConfigurableBase):
 
     def __len__(self):
         self.log.trace()
-        return len(self.msgids)
+        return len(self.msgnum_by_msgid)
 
     def __getitem__(self, i):
         self.log.trace('i == %d' % i)
         if not self.__initialized:
             raise getmailOperationError('not initialized')
-        return self.msgids[i]
+        return sorted(self.msgid_by_msgnum.items())[i][1]
 
     def _read_oldmailfile(self):
         '''Read contents of oldmail file.'''
@@ -391,7 +358,7 @@ class RetrieverSkeleton(ConfigurableBase):
             # Explicitly set to None in case it was previously set
             socket.setdefaulttimeout(None)
         self.oldmail_filename = os.path.join(
-            expand_user_vars(self.conf['getmaildir']),
+            self.conf['getmaildir'],
             ('oldmail-%(server)s-%(port)i-%(username)s' % self.conf).replace(
                 '/', '-').replace(':', '-')
         )
@@ -442,9 +409,9 @@ class POP3RetrieverBase(RetrieverSkeleton):
 
     def _getmsgnumbyid(self, msgid):
         self.log.trace()
-        if not msgid in self.msgids:
+        if not msgid in self.msgnum_by_msgid:
             raise getmailOperationError('no such message ID %s' % msgid)
-        return self.msgids.index(msgid) + 1
+        return self.msgnum_by_msgid[msgid]
 
     def _getmsglist(self):
         self.log.trace()
@@ -452,21 +419,43 @@ class POP3RetrieverBase(RetrieverSkeleton):
             response, msglist, octets = self.conn.uidl()
             self.log.debug('UIDL response "%s", %d octets'
                 % (response, octets) + os.linesep)
-            for msgid in [line.split(None, 1)[1] for line in msglist]:
-                if msgid in self.msgids:
+            for (i, line) in enumerate(msglist):
+                try:
+                    (msgnum, msgid) = line.split(None, 1)
+                except ValueError:
+                    # Line didn't contain two tokens.  Server is broken.
                     raise getmailOperationError(
-                        '%s does not uniquely identify messages '
-                        '(got %s twice) -- use BrokenUIDLPOP3Retriever instead'
-                        % (self, msgid)
+                        '%s failed to identify message index %d in UIDL output'
+                        ' -- see documentation or use '
+                        'BrokenUIDLPOP3Retriever instead'
+                        % (self, i)
                     )
+                msgnum = int(msgnum)
+                if msgid in self.msgnum_by_msgid:
+                    # UIDL "unique" identifiers weren't unique.  
+                    # Server is broken.
+                    if self.conf.get('delete_dup_msgids', False):
+                        self.log.debug('deleting message %s with duplicate '
+                                       'msgid %s' % (msgnum, msgid) 
+                                       + os.linesep)
+                        self.conn.dele(msgnum)
+                    else:
+                        raise getmailOperationError(
+                            '%s does not uniquely identify messages '
+                            '(got %s twice) -- see documentation or use '
+                            'BrokenUIDLPOP3Retriever instead'
+                            % (self, msgid)
+                        )
                 else:
-                    self.msgids.append(msgid)
-            self.log.debug('Message IDs: %s' % self.msgids + os.linesep)
+                    self.msgnum_by_msgid[msgid] = msgnum
+                    self.msgid_by_msgnum[msgnum] = msgid
+            self.log.debug('Message IDs: %s' 
+                           % sorted(self.msgnum_by_msgid.keys()) + os.linesep)
             response, msglist, octets = self.conn.list()
             for line in msglist:
                 msgnum = int(line.split()[0])
                 msgsize = int(line.split()[1])
-                self.msgsizes[self.msgids[msgnum - 1]] = msgsize
+                self.msgsizes[self.msgid_by_msgnum[msgnum]] = msgsize
         except poplib.error_proto, o:
             raise getmailOperationError(
                 'POP error (%s) - if your server does not support the UIDL '
@@ -514,7 +503,8 @@ class POP3RetrieverBase(RetrieverSkeleton):
                 self.conn.user(self.conf['username'])
                 self.conn.pass_(self.conf['password'])
             self._getmsglist()
-            self.log.debug('msgids: %s' % self.msgids + os.linesep)
+            self.log.debug('msgids: %s' 
+                           % sorted(self.msgnum_by_msgid.keys()) + os.linesep)
             self.log.debug('msgsizes: %s' % self.msgsizes + os.linesep)
             # Remove messages from state file that are no longer in mailbox
             for msgid in self.oldmail.keys():
@@ -615,7 +605,7 @@ class IMAPRetrieverBase(RetrieverSkeleton):
 
     def _getmboxuidbymsgid(self, msgid):
         self.log.trace()
-        if not msgid in self.msgids:
+        if not msgid in self.msgnum_by_msgid:
             raise getmailOperationError('no such message ID %s' % msgid)
         mailbox, uid = self._mboxuids[msgid]
         return mailbox, uid
@@ -655,18 +645,23 @@ class IMAPRetrieverBase(RetrieverSkeleton):
         r = {}
         try:
             parts = line[line.index('(') + 1:line.rindex(')')].split()
-            if len(parts) % 2:
-                # Not zero -- i.e. an uneven number of parts
-                raise ValueError
             while parts:
-                # Interesting; RHS is evaluated first, breaking this
-                # expression...
-                # r[parts.pop(0).lower()] = parts.pop(0)
+                # Flags starts a parenthetical list of valueless flags
+                if parts[0].lower() == 'flags' and parts[1].startswith('('):
+                    while parts and not parts[0].endswith(')'):
+                        del parts[0]
+                    if parts:
+                        # Last one, ends with ")"
+                        del parts[0]
+                    continue
+                if len(parts) == 1:
+                    # Leftover part -- not name, value pair.
+                    raise ValueError
                 name = parts.pop(0).lower()
                 r[name] = parts.pop(0)
-        except ValueError:
+        except (ValueError, IndexError), o:
             raise getmailOperationError('IMAP error (failed to parse'
-                ' UID response line %s)' % line)
+                ' UID response line "%s")' % line)
         self.log.trace('got %s' % r + os.linesep)
         return r
 
@@ -692,8 +687,9 @@ class IMAPRetrieverBase(RetrieverSkeleton):
 
     def _getmsglist(self):
         self.log.trace()
-        self.msgids = []
+        self.msgnum_by_msgid = {}
         self._mboxuids = {}
+        self._mboxuidorder = []
         self.msgsizes = {}
         for mailbox in self.conf['mailboxes']:
             try:
@@ -708,12 +704,16 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                         msgid = ('%s/%s/%s'
                             % (self.uidvalidity, mailbox, r['uid']))
                         self._mboxuids[msgid] = (mailbox, r['uid'])
-                        self.msgids.append(msgid)
+                        self._mboxuidorder.append(msgid)
+                        self.msgnum_by_msgid[msgid] = None
                         self.msgsizes[msgid] = int(r['rfc822.size'])
             except imaplib.IMAP4.error, o:
                 raise getmailOperationError('IMAP error (%s)' % o)
         self.gotmsglist = True
 
+    def __getitem__(self, i):
+        return self._mboxuidorder[i]
+    
     def _delmsgbyid(self, msgid):
         self.log.trace()
         try:
@@ -783,7 +783,8 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                     self.conf['password'])
             self.log.trace('logged in, getting message list' + os.linesep)
             self._getmsglist()
-            self.log.debug('msgids: %s' % self.msgids + os.linesep)
+            self.log.debug('msgids: %s' 
+                           % sorted(self.msgnum_by_msgid.keys()) + os.linesep)
             self.log.debug('msgsizes: %s' % self.msgsizes + os.linesep)
             # Remove messages from state file that are no longer in mailbox
             for msgid in self.oldmail.keys():
@@ -868,6 +869,7 @@ class MultidropIMAPRetrieverBase(IMAPRetrieverBase):
         return msg
 
 
+# Choose right POP-over-SSL mix-in based on Python version being used.
 if sys.hexversion >= 0x02040000:
     POP3SSLinitMixIn = Py24POP3SSLinitMixIn
 else:
