@@ -31,9 +31,9 @@ import email.Utils
 
 import pwd
 
-from exceptions import *
-from utilities import *
-from baseclasses import ConfigurableBase, ForkingBase
+from getmailcore.exceptions import *
+from getmailcore.utilities import *
+from getmailcore.baseclasses import *
 
 #######################################
 class DeliverySkeleton(ConfigurableBase):
@@ -104,19 +104,21 @@ class Maildir(DeliverySkeleton, ForkingBase):
       '~USER/', as well as environment variables.
     '''
     _confitems = (
-        {'name' : 'path', 'type' : str},
-        {'name' : 'configparser', 'type' : types.InstanceType, 'default' : None},
-        {'name' : 'user', 'type' : str, 'default' : None},
+        ConfInstance(name='configparser', required=False),
+        ConfMaildirPath(name='path'),
+        ConfString(name='user', required=False, default=None),
+        ConfString(name='filemode', required=False, default='0600'),
     )
 
     def initialize(self):
         self.log.trace()
         self.hostname = localhostname()
         self.dcount = 0
-        self.conf['path'] = expand_user_vars(self.conf['path'])
-        if not self.conf['path'].endswith('/'):
-            raise getmailConfigurationError('maildir path missing trailing /'
-                ' (%s)' % self.conf['path'])
+        try:
+            self.conf['filemode'] = int(self.conf['filemode'], 8)
+        except ValueError, o:
+            raise getmailConfigurationError('filemode %s not valid: %s'
+                % (self.conf['filemode'], o))
 
     def __str__(self):
         self.log.trace()
@@ -142,7 +144,7 @@ class Maildir(DeliverySkeleton, ForkingBase):
                         'refuse to deliver mail as GID 0'
                     )
             f = deliver_maildir(self.conf['path'], msg.flatten(delivered_to,
-                received), self.hostname, self.dcount)
+                received), self.hostname, self.dcount, self.conf['filemode'])
             stdout.write(f)
             stdout.flush()
             os.fsync(stdout.fileno())
@@ -218,14 +220,19 @@ class Mboxrd(DeliverySkeleton, ForkingBase):
     http://groups.google.com/groups?selm=4ivk9s%24bok%40hustle.rahul.net
     '''
     _confitems = (
-        {'name' : 'path', 'type' : str},
-        {'name' : 'configparser', 'type' : types.InstanceType, 'default' : None},
-        {'name' : 'user', 'type' : str, 'default' : None},
+        ConfInstance(name='configparser', required=False),
+        ConfMboxPath(name='path'),
+        ConfString(name='locktype', required=False, default='lockf'),
+        ConfString(name='user', required=False, default=None),
     )
 
     def initialize(self):
         self.log.trace()
-        self.conf['path'] = expand_user_vars(self.conf['path'])
+        if self.conf['locktype'] not in ('lockf', 'flock'):
+            raise getmailConfigurationError(
+                'unknown mbox lock type: %s'
+                % self.conf['locktype']
+            )
 
     def __str__(self):
         self.log.trace()
@@ -262,14 +269,14 @@ class Mboxrd(DeliverySkeleton, ForkingBase):
             fd = os.open(self.conf['path'], os.O_RDWR)
             status_old = os.fstat(fd)
             f = os.fdopen(fd, 'r+b')
-            lock_file(f)
+            lock_file(f, self.conf['locktype'])
             # Check if it _is_ an mbox file.  mbox files must start with "From "
             # in their first line, or are 0-length files.
             f.seek(0, 0)
             first_line = f.readline()
-            if first_line and first_line[:5] != 'From ':
+            if first_line and not first_line.startswith('From '):
                 # Not an mbox file; abort here
-                unlock_file(f)
+                unlock_file(f, self.conf['locktype'])
                 raise getmailDeliveryError('not an mboxrd file (%s)'
                     % self.conf['path'])
             # Seek to end
@@ -298,7 +305,7 @@ class Mboxrd(DeliverySkeleton, ForkingBase):
                     stdout.flush()
                     os.fsync(stdout.fileno())
 
-                unlock_file(f)
+                unlock_file(f, self.conf['locktype'])
 
             except IOError, o:
                 try:
@@ -439,26 +446,26 @@ class MDA_qmaillocal(DeliverySkeleton, ForkingBase):
     '''
 
     _confitems = (
-        {'name' : 'qmaillocal', 'type' : str, 'default' : '/var/qmail/bin/qmail-local'},
-        {'name' : 'user', 'type' : str, 'default' : pwd.getpwuid(os.geteuid()).pw_name},
-        {'name' : 'group', 'type' : str, 'default' : None},
-        {'name' : 'homedir', 'type' : str, 'default' : pwd.getpwuid(os.geteuid()).pw_dir},
-        {'name' : 'localdomain', 'type' : str, 'default' : localhostname()},
-        {'name' : 'defaultdelivery', 'type' : str, 'default' : './Maildir/'},
-        {'name' : 'conf-break', 'type' : str, 'default' : '-'},
-        {'name' : 'localpart_translate', 'type' : tuple, 'default' : ('', '')},
-        {'name' : 'strip_delivered_to', 'type' : bool, 'default' : False},
-        {'name' : 'allow_root_commands', 'type' : bool, 'default' : False},
-        {'name' : 'configparser', 'type' : types.InstanceType, 'default' : None},
+        ConfInstance(name='configparser', required=False),
+        ConfFile(name='qmaillocal', required=False, 
+                 default='/var/qmail/bin/qmail-local'),
+        ConfString(name='user', required=False, 
+                   default=pwd.getpwuid(os.geteuid()).pw_name),
+        ConfString(name='group', required=False, default=None),
+        ConfDirectory(name='homedir', required=False, 
+                      default=pwd.getpwuid(os.geteuid()).pw_dir),
+        ConfString(name='localdomain', required=False, default=localhostname()),
+        ConfString(name='defaultdelivery', required=False, 
+                   default='./Maildir/'),
+        ConfString(name='conf-break', required=False, default='-'),
+        ConfTupleOfStrings(name='localpart_translate', required=False, 
+                           default="('', '')"),
+        ConfBool(name='strip_delivered_to', required=False, default=False),
+        ConfBool(name='allow_root_commands', required=False, default=False),
     )
 
     def initialize(self):
         self.log.trace()
-        self.conf['qmaillocal'] = expand_user_vars(self.conf['qmaillocal'])
-        self.conf['homedir'] = expand_user_vars(self.conf['homedir'])
-        if not os.path.isdir(self.conf['homedir']):
-            raise getmailConfigurationError('no such directory %s'
-                % self.conf['homedir'])
 
     def __str__(self):
         self.log.trace()
@@ -478,7 +485,9 @@ class MDA_qmaillocal(DeliverySkeleton, ForkingBase):
             self.log.debug('about to execl() with args %s\n' % str(args))
             # Modify message
             if self.conf['strip_delivered_to']:
-                del msg['delivered-to']
+                msg.remove_header('delivered-to')
+                # Also don't insert a Delivered-To: header.
+                delivered_to = None
             # Write out message
             msgfile = os.tmpfile()
             msgfile.write(msg.flatten(delivered_to, received))
@@ -496,7 +505,9 @@ class MDA_qmaillocal(DeliverySkeleton, ForkingBase):
             if ((os.geteuid() == 0 or os.getegid() == 0)
                     and not self.conf['allow_root_commands']):
                 raise getmailConfigurationError(
-                    'refuse to invoke external commands as root or GID 0 by default')
+                    'refuse to invoke external commands as root '
+                    'or GID 0 by default'
+                )
 
             os.execl(*args)
         except StandardError, o:
@@ -565,11 +576,16 @@ class MDA_qmaillocal(DeliverySkeleton, ForkingBase):
         if exitcode == 111:
             raise getmailDeliveryError('qmail-local %d temporary error (%s)'
                 % (childpid, err))
-        elif exitcode or err:
+        elif exitcode:
             raise getmailDeliveryError('qmail-local %d error (%d, %s)'
                 % (childpid, exitcode, err))
 
-        return 'MDA_qmaillocal (%s)' % out
+        if out and err:
+            info = '%s:%s' % (out, err)
+        else:
+            info = out or err
+
+        return 'MDA_qmaillocal (%s)' % info
 
 #######################################
 class MDA_external(DeliverySkeleton, ForkingBase):
@@ -613,24 +629,24 @@ class MDA_external(DeliverySkeleton, ForkingBase):
       allow_root_commands (boolean, optional) - if set, external commands are
             allowed when running as root.  The default is not to allow such
             behaviour.
+    
+      ignore_stderr (boolean, optional) - if set, getmail will not consider the
+            program writing to stderr to be an error.  The default is False.
     '''
     _confitems = (
-        {'name' : 'path', 'type' : str},
-        {'name' : 'arguments', 'type' : tuple, 'default' : ()},
-        {'name' : 'unixfrom', 'type' : bool, 'default' : False},
-        {'name' : 'user', 'type' : str, 'default' : None},
-        {'name' : 'group', 'type' : str, 'default' : None},
-        {'name' : 'allow_root_commands', 'type' : bool, 'default' : False},
-        {'name' : 'configparser', 'type' : types.InstanceType, 'default' : None},
+        ConfInstance(name='configparser', required=False),
+        ConfFile(name='path'),
+        ConfTupleOfStrings(name='arguments', required=False, default="()"),
+        ConfString(name='user', required=False, default=None),
+        ConfString(name='group', required=False, default=None),
+        ConfBool(name='allow_root_commands', required=False, default=False),
+        ConfBool(name='unixfrom', required=False, default=False),
+        ConfBool(name='ignore_stderr', required=False, default=False),
     )
 
     def initialize(self):
         self.log.trace()
-        self.conf['path'] = expand_user_vars(self.conf['path'])
         self.conf['command'] = os.path.basename(self.conf['path'])
-        if not os.path.isfile(self.conf['path']):
-            raise getmailConfigurationError('no such command %s'
-                % self.conf['path'])
         if not os.access(self.conf['path'], os.X_OK):
             raise getmailConfigurationError('%s not executable'
                 % self.conf['path'])
@@ -667,7 +683,9 @@ class MDA_external(DeliverySkeleton, ForkingBase):
             if ((os.geteuid() == 0 or os.getegid() == 0)
                     and not self.conf['allow_root_commands']):
                 raise getmailConfigurationError(
-                    'refuse to invoke external commands as root or GID 0 by default')
+                    'refuse to invoke external commands as root '
+                    'or GID 0 by default'
+                )
             args = [self.conf['path'], self.conf['path']]
             for arg in self.conf['arguments']:
                 arg = expand_user_vars(arg)
@@ -717,9 +735,12 @@ class MDA_external(DeliverySkeleton, ForkingBase):
         self.log.debug('command %s %d exited %d\n' % (self.conf['command'],
             childpid, exitcode))
 
-        if exitcode or err:
+        if exitcode or (err and not self.conf['ignore_stderr']):
             raise getmailDeliveryError('command %s %d error (%d, %s)'
                 % (self.conf['command'], childpid, exitcode, err))
+        elif err:
+            # User said to ignore stderr, just log it.
+            self.log.info('command %s: %s' % (self, err))
 
         return 'MDA_external command %s (%s)' % (self.conf['command'], out)
 
@@ -732,7 +753,7 @@ class MultiDestinationBase(DeliverySkeleton):
 
       conf - standard ConfigurableBase configuration dictionary
 
-      log - getmailcore.logging.logger() instance
+      log - getmailcore.logging.Logger() instance
 
     In addition, sub-classes must populate the following list provided by
     this base class:
@@ -801,8 +822,8 @@ class MultiDestination(MultiDestinationBase):
                 then interpreted as maildir/mbox/other-destination-section.
     '''
     _confitems = (
-        {'name' : 'destinations', 'type' : tuple},
-        {'name' : 'configparser', 'type' : types.InstanceType, 'default' : None},
+        ConfInstance(name='configparser', required=False),
+        ConfTupleOfStrings(name='destinations'),
     )
 
     def initialize(self):
@@ -871,8 +892,7 @@ class MultiSorterBase(MultiDestinationBase):
                 except getmailConfigurationError, o:
                     raise getmailConfigurationError('pattern %s destination'
                         ' error %s' % (pattern, o))
-                self.targets.append( (re.compile(pattern.replace('\\', '\\\\'),
-                    re.IGNORECASE), dest) )
+                self.targets.append((re.compile(pattern, re.IGNORECASE), dest))
                 self._destinations.append(dest)
         except re.error, o:
             raise getmailConfigurationError('invalid regular expression %s' % o)
@@ -929,9 +949,9 @@ class MultiSorter(MultiSorterBase):
                with them.
     '''
     _confitems = (
-        {'name' : 'default', 'type' : str},
-        {'name' : 'locals', 'type' : tuple, 'default' : ()},
-        {'name' : 'configparser', 'type' : types.InstanceType, 'default' : None},
+        ConfInstance(name='configparser', required=False),
+        ConfString(name='default'),
+        ConfTupleOfTupleOfStrings(name='locals', required=False, default="()"),
     )
 
     def __str__(self):
@@ -978,9 +998,9 @@ class MultiGuesser(MultiSorterBase):
 
     '''
     _confitems = (
-        {'name' : 'default', 'type' : str},
-        {'name' : 'locals', 'type' : tuple, 'default' : ()},
-        {'name' : 'configparser', 'type' : types.InstanceType, 'default' : None},
+        ConfInstance(name='configparser', required=False),
+        ConfString(name='default'),
+        ConfTupleOfTupleOfStrings(name='locals', required=False, default="()"),
     )
 
     def __str__(self):
